# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Load and filter files from the Git repository or cache."""


def dotsmug_paths(path):
    """Make a list of the .smug files that should be loaded for a given path.

    >>> dotsmug_paths('path/to/file')
    ['path/to/.smug', 'path/.smug', '.smug']
    >>> dotsmug_paths('path/to/dir/')
    ['path/to/dir/.smug', 'path/to/.smug', 'path/.smug', '.smug']
    >>>
    """
    path_splits = path.split('/')[:-1]

    subpaths = ['']
    for split in path_splits:
        next = subpaths[-1] + split + '/'
        subpaths.append(next)

    results = [(subpath + '.smug') for subpath in subpaths]
    results.reverse()

    return results


def load_dotsmug(request_path, repo):
    """Load the .smug config file as a dotsmug.Config."""

    from dotsmug import DotSmug, Config

    paths = dotsmug_paths(request_path)
    rawfiles = [SourceFile(path, 'master', repo) for path in paths]
    dotsmugs = [DotSmug(raw.content) for raw in rawfiles if raw.exists()]

    config = Config(dotsmugs)
    return config


def get_source(path, branch, repo):
    """Gets the source file and filter for the given request path.
    
    The .smug settings are used to determine the source path from the request
    path.

    Returns (source_file, filter, filter_args).
    """
    # Load configuration:
    dotsmug = load_dotsmug(path, repo)

    if not path or path.endswith('/'):
        index = dotsmug.directory_index()
        if index:
            path += index

    for source_path, using in dotsmug.convert_sources(path):
        source_file = SourceFile(source_path, branch, repo)
        if source_file.exists():
            from filters import get_filter
            filter = get_filter(using)
            break
    else:
        source_file = SourceFile(path, branch, repo)
        source_path = path
        filter = None

    filter_args = {}
    template = dotsmug.template()
    if template:
        filter_args['template'] = template

    return source_file, filter, filter_args


def get_content(request_path, branch, repo, raw=False, **filter_args):
    """Gets a file from the Git repo, using a cache if possible."""
    if raw:
        dest_file = SourceFile(request_path, branch, repo)
    else:
        dest_file = FilteredFile(request_path, branch, repo, filter_args)

    if not dest_file.exists():
        from django.http import Http404
        raise Http404

    return dest_file


# TODO: Update merging, etc., to update the head ref as they write to the repo.
class HeadCache(object):
    """Accesses and updates the cached reference to the head of a branch."""

    def __init__(self, repo, branch='master'):
        self.key = 'head:%s(%s)' % (repo.path, branch)
        self.repo = repo
        self.branch = branch

    def get(self):
        """Gets the commit id of the head of the given branch."""
        from django.core.cache import cache
        head = cache.get(self.key)
        if head is None:
            #print 'cache miss for head:', self.key
            head = self.update()
        return head

    def update(self, head=None):
        """Update the cached head ref from a specified value or from disk.

        Note that it's possible to accidentally set the head ref to an older
        value, but the head ref will never be more than than `head_timeout`
        seconds old.
        """
        import config
        from django.core.cache import cache
        if head is None:
            head = self.repo.find_head(self.branch)
        cache.set(self.key, head, config.head_timeout)
        return head


def cached_property(name):
    """A property that forces a self.lookup() call when read.
    
    The given name is the name of the cached variable (usually starting with
    an underscore).
    """
    def getter(self):
        if self.delayed:
            self.lookup()
        return getattr(self, name)
    def setter(self, value):
        setattr(self, name, value)
    return property(getter, setter)


class CachedFile(object):
    """Raw or filtered content.

    Smug will try to cache data as much as possible, so all requests to the
    repository should try to go through CachedFile.
    """

    KEY_PREFIX = None

    def __init__(self, path, branch, repo):
        self.path = path
        self.branch = branch
        self.repo = repo
        self.key = '%s:%s//%s(%s)' % (self.KEY_PREFIX, repo.path, path, branch)
        self.head_cache = HeadCache(repo, branch)

        # If delayed is set, information has not yet been retrieved/generated.
        self.delayed = True

        self._commit = None
        self._type = None
        self._mimetype = None
        self._content = None
        self._sourcename = None
        self._sourcepath = None

    def is_directory(self):
        if self.delayed:
            self.lookup()
        return self._type == 'tree'

    def exists(self):
        if self.delayed:
            self.lookup()
        return self._type is not None

    def source_file(self):
        raise NotImplementedError

    def lookup(self):
        """Try to find all information, either from the cache or from Git."""
        from django.core.cache import cache
        values = cache.get(self.key)
        self.delayed = False

        if values:
            (self.commit, self.type, self.mimetype, self.content,
                    self.sourcename, self.sourcepath) = values
            head = self.head_cache.get()
            if self.commit == head:
                #print 'cache hit for:', self.key
                loaded = True
            else:
                # The cache is out of date but might still be correct.
                #print 'cache miss (outdated commit) for:', self.key
                self.commit = None
                loaded = False
        else:
            # Cache miss
            #print 'cache miss for:', self.key
            (self.commit, self.type, self.mimetype, self.content,
                    self.sourcename, self.sourcepath) = (None,) * 6
            loaded = False

        if not loaded:
            self.retrieve()
            # TODO: only cache on success?
            self.cache_data()

    def cache_data(self):
        """Save data to the cache."""
        from django.core.cache import cache
        values = (self._commit, self._type, self._mimetype, self._content,
                self._sourcename, self._sourcepath)
        cache.set(self.key, values)

    def retrieve(self):
        """Try to retrieve content from the actual source.

        This must be implemented by subclasses.
        """
        raise NotImplementedError

    def dirty(self):
        """Clear information from the cache."""
        #print 'trying to dirty cache for "%s"' % self.key
        from django.core.cache import cache
        cache.delete(self.key)
        self.delayed = True

    commit = cached_property('_commit')
    content = cached_property('_content')
    type = cached_property('_type')
    mimetype = cached_property('_mimetype')
    sourcename = cached_property('_sourcename')
    sourcepath = cached_property('_sourcepath')


class SourceFile(CachedFile):
    """A file from a Git repository."""

    KEY_PREFIX = 'source'

    def source_file(self):
        return self

    def retrieve(self):
        """Try to retrieve content from the Git repository.

        If there was a complete cache miss, the object properties (commit,
        sourcename, etc.) will be None before retrieve is called.  If there
        was an outdated cache hit, then self.commit will be None but the rest
        of the properties will be set to the cached values.
        """
        commit = self.head_cache.get()
        name = self.repo.getname(self.path, treename=commit)
        if not name:
            # TODO: cache nonexistent files (?)
            return

        # Check for an outdated but valid cache hit.
        if not self.commit and self.sourcename and (self.sourcename == name):
            self.commit = commit
            return

        self.sourcename = name
        self.commit = commit
        self.type = self.repo.gettype(name)
        if self.type == 'blob':
            blob = self.repo.getblob(name)
            self.content = blob.contents
            self.sourcepath = self.path

            from mimetypes import guess_type
            self.mimetype, encoding = guess_type(self.path, strict=False)


class FilteredFile(CachedFile):
    """Filtered content.

    For now, this will handle all filtered content.  In the future, it will
    only handle Djangoized content.
    """

    KEY_PREFIX = 'filtered'

    def __init__(self, path, branch, repo, filter_args):
        CachedFile.__init__(self, path, branch, repo)
        self.filter_args = filter_args

    def source_file(self):
        return SourceFile(self.sourcepath, self.branch, self.repo)

    def retrieve(self):
        source_file, filter, filter_args = get_source(self.path, self.branch,
                self.repo)
        if not source_file.exists():
            return

        # Check for an outdated but valid cache hit.
        if not self.commit and self.sourcename:
            if (self.sourcename == source_file.sourcename):
                self.commit = source_file.commit
                return

        self.commit = source_file.commit
        self.content = source_file.content
        self.type = source_file.type
        self.mimetype = source_file.mimetype
        self.sourcepath = source_file.sourcepath

        # Guess mimetype for extensionless files before sending on to filter.
        # Note that we don't do this in SourceFile.retrieve because then
        # load_dotsmug would get in an infinite loop.  The open question
        # though, is whether it is more appropriate to set the mimetype based
        # on self.path or self.sourcepath.
        if not self.mimetype:
            dotsmug = load_dotsmug(self.path, self.repo)
            default_mimetype = dotsmug.default_mimetype()
            if default_mimetype:
                self.mimetype = default_mimetype
            else:
                import config
                self.mimetype = config.DEFAULT_MIMETYPE

        if filter:
            # TODO: allow filters to be chained (and only cache the last one)
            kwds = self.filter_args.copy()
            kwds.update(filter_args)
            filter(self, repo=self.repo, **kwds)


if __name__ == '__main__':
    import doctest
    doctest.testmod()

# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


class Config(object):
    """Process configuration queries using a hierarchy of .smug files.
    
    The DotSmug objects are processed in the order in which they are passed
    in.  Thus, the innermost DotSmug should be first in the list.
    """
    def __init__(self, dotsmugs):
        self.dotsmugs = dotsmugs

    def directory_index(self):
        """Find the directory index."""
        for dotsmug in self.dotsmugs:
            index = dotsmug.directory_index
            if index is not None:
                return index
        return None

    def template(self):
        """Find the default template"""
        for dotsmug in self.dotsmugs:
            template = dotsmug.template
            if template is not None:
                return template
        return None

    def default_mimetype(self):
        """Find the default mimetype."""
        for dotsmug in self.dotsmugs:
            mimetype = dotsmug.default_mimetype
            if mimetype is not None:
                return mimetype
        return None

    def convert_sources(self, path):
        """Iterate over convert commands which apply to the given request.

        Yields source_path and using.
        """
        for dotsmug in self.dotsmugs:
            for convert_line in dotsmug.convert:
                convert_to = convert_line['to']
                if path.endswith(convert_to):
                    convert_from = convert_line['from']
                    convert_using = convert_line['using']
                    base = path[0:-len(convert_to)]
                    source_path = base + convert_from

                    # in the future, a SMUG_DEBUG option should make this print happen:
                    #print 'request: %s, source: %s, using: %s' % (path, source_path, convert_using)

                    yield source_path, convert_using


class DotSmugError(Exception):
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return 'Error in .smug: %s' % self.text

class DotSmug(object):
    """Parse a .smug file.
    
    The directory_index directive sets the directory_index attribute:
    >>> teststr = "directory_index abc"
    >>> config = DotSmug(teststr)
    >>> config.directory_index
    'abc'
    >>>

    The directory_index directive may only be specified once:
    >>> teststr = '''directory_index abc
    ... directory_index extra_line'''
    >>> try: config = DotSmug(teststr)
    ... except DotSmugError:
    ...   print 'failed'
    failed
    >>>

    Exactly one argument must be given for directory_index:
    >>> teststr = "directory_index abc extra"
    >>> try: config = DotSmug(teststr)
    ... except DotSmugError:
    ...   print 'failed'
    failed
    >>>

    """

    def __init__(self, dotsmugstr):
        self.tokenizer = line_tokenizer(dotsmugstr)

        self.directory_index = None
        self.default_mimetype = None
        self.template = None
        self.convert = []

        self.parse()

    def parse(self):
        """Parse the .smug file."""

        valid_directives = ('convert', 'directory_index', 'default_mimetype','template')

        for line in self.tokenizer:
            if not line:
                continue
            directive = line[0]
            if directive in valid_directives:
                subparser_name = 'parse_%s' % directive
                subparser = getattr(self, subparser_name)
                subparser(line)
            else:
                raise DotSmugError('unknown directive in .smug file.')

    def parse_convert(self, line):
        """Parse a convert directive."""

        valid_arguments = ('from', 'to', 'using')
        args = dict([(argname, None) for argname in valid_arguments])

        it = iter(line[1:])
        while True:
            try:
                argname = it.next()
            except StopIteration:
                break

            if argname not in args:
                raise DotSmugError('invalid argument name: %s.' % argname)

            if args[argname] is not None:
                raise DotSmugError('argument "%s" specified more than once.'
                        % argname)

            try:
                argvalue = it.next()
            except StopIteration:
                raise DotSmugError('no value for argument "%s".' % argname)

            args[argname] = argvalue

        # default value should be empty string
        for argname in valid_arguments:
            if args[argname] is None:
                args[argname] = ''

        self.convert.append(args)

    def parse_directory_index(self, line):
        """Parse the directory_index directive."""

        if self.directory_index is not None:
            raise DotSmugError('directory_index specified more than once.')
        if len(line) != 2:
            raise DotSmugError('wrong number of arguments to directory_index.')

        self.directory_index = line[1]

    def parse_template(self, line):
        """Parses the template directive."""

        if self.template is not None:
            raise DotSmugError('template specified more than once.')
        if len(line) != 2:
            raise DotSmugError('wrong number of arguments to template.')
        self.template = line[1]

    def parse_default_mimetype(self, line):
        """Parse the default_mimetype directive."""

        if self.default_mimetype is not None:
            raise DotSmugError('default_mimetype specified more than once.')
        if len(line) != 2:
            raise DotSmugError('wrong number of arguments to default_mimetype.')

        self.default_mimetype = line[1]


def line_tokenizer(inputstring):
    r"""Yield a list of tokens for each line in a file.

    If a quoted string is incomplete at the end of a physical line, the
    logical line continues to the next line.  Likewise, join two lines if
    the first line ends in a backslash.

    Setup a test file:
    >>> teststr = r'''First line.
    ... Beginning of second \
    ... line and end.
    ... Beginning of "third
    ... line" and end.'''
    >>>

    Make sure that the lexical analyzer works:
    >>> it = line_tokenizer(teststr)
    >>> it.next()
    ['First', 'line.']
    >>> it.next()
    ['Beginning', 'of', 'second', 'line', 'and', 'end.']
    >>> it.next()
    ['Beginning', 'of', 'third\nline', 'and', 'end.']
    >>>
    """
    from shlex import split

    logical_line = ''
    tokens = []
    for physical_line in inputstring.splitlines(True):
        logical_line += physical_line

        try:
            new_tokens = split(logical_line, comments=True)
        except ValueError:
            # Unmatched quote at end of line: grab another physical line.
            continue

        if new_tokens and new_tokens[-1] == '\n':
            # Continuation (line ends in backslash)
            tokens += new_tokens[:-1]
            # Grab another logical line.
            logical_line = ''
            continue
        else:
            # complete line
            tokens += new_tokens
            yield tokens

        logical_line = ''
        tokens = []

    if logical_line:
        # End of file has been reached but logical_line has not been cleared.
        raise ValueError("Syntax error: unmatched quote at EOF")


def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()

# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.shortcuts import render_to_response
from django.contrib.auth.decorators import login_required
from django import forms

from smug import config
from smug.load import get_source, HeadCache


class AnonymousEditForm(forms.Form):
    """Edit form for anonymous users."""
    text = forms.CharField(widget=forms.widgets.Textarea(
            attrs={'cols': 80, 'rows': 24}))
    changelog = forms.CharField(widget=forms.widgets.Textarea(
            attrs={'cols': 80, 'rows': 3}))
    parent = forms.CharField(widget=forms.widgets.HiddenInput)
    noeol = forms.CharField(widget=forms.widgets.HiddenInput)


class AuthenticatedEditForm(AnonymousEditForm):
    """Edit form for authenticated users.
    
    Unlike anonymous users, authenticated users can pick a branch to commit to.
    """

    # TODO: make a radio button: "commit in place" vs. "start new branch."
    # They can only commit in place if they have write-access to the branch
    # (i.e. they're either an editor or they own the branch).  Maybe even
    # give the option to automatically pick the name for the new branch.
    branch = forms.CharField(label='Commit to Branch')


def edit(request, page, repo, branch, **kwds):
    """Edit a page in the wiki.

    This function both displays the form and processes its results.
    """
    source_file, filter, filter_args  = get_source(page, branch, repo)
    path = source_file.sourcepath
    if not path:
        path = page

    user = request.user

    if request.method == 'POST':
        text = request.POST['text'].encode('utf-8')
        changelog = request.POST['changelog'].encode('utf-8')
        parent = request.POST['parent']
        if request.POST['noeol'] == 'False':
            text = text + '\n'

        # Normalize whitespace--use Unix newlines:
        text = text.replace('\r\n', '\n')

        if user.is_authenticated():
            # Get branch and parent from form
            newbranch = request.POST['branch']

            success = commit_auth(user, path, parent, changelog, text,
                    newbranch, repo)

            if not success:
                # TODO: Do something better than just return.  Perhaps we
                # should try to do a fast-forward merge or recommend that they
                # create a new branch and merge or something else completely.
                # Or maybe the commit succeeded earlier, and they just
                # reposted it.
                return HttpResponse('Commit failed!')
        else:
            commit_anon(path, branch, changelog, text)

        if newbranch == 'master':
            redirect = request.path
        else:
            redirect = '%s?branch=%s' % (request.path, newbranch)
        return HttpResponseRedirect(redirect)
    else:
        # Supply a page with an edit form
        return edit_form(path, branch, repo, user)


def commit_auth(user, path, parent, changelog, text, newbranch, repo):
    """Commit changes as an authenticated user."""

    # Set name and email for Git commit
    if user.first_name and user.last_name:
        author_name = ' '.join((user.first_name, user.last_name))
    else:
        author_name = user.username
    if user.email:
        author_email = user.email
    else:
        author_email = None

    # TODO: it's probably smart to do a fast-forward merge here when the
    # branch has moved ahead of us and parent != currenthead.  Here's how that
    # looks:
    # 1) do git-add on the changes that the person made
    # 2) do a "git-merge -m parent -m newhead"
    # 3) do a commit and save_head
    # If the save_head command fails, maybe the smart thing to do is to try to
    # repeat this whole process a few times in a row.

    newcommit = repo.commit_add(path, text, parent, changelog,
            author_name, author_email)

    # This will fail if the new commit is not a child of the current head or
    # if the current head changes on us while we are committing:
    success = repo.save_head(newcommit, parent, newbranch)
    if success:
        head_cache = HeadCache(repo, newbranch)
        head_cache.update(newcommit)

    return success


def commit_anon(path, parent, changelog, text):
    """Commit changes as an anonymous user."""

    from difflib import unified_diff
    from smug.models import Patch

    # Retrieve old text.
    name = repo.getname(path, treename=parent)
    blob = repo.getblob(name)
    if blob is None:
        # TODO: Fail here in a better way.
        raise Http404

    patch = Patch()
    patch.status = 's'
    patch.parent = parent
    patch.changelog = changelog

    fromfile = '%s.%s' % (path, parent)
    udiff = unified_diff(blob.contents.splitlines(), text.splitlines(),
            fromfile=fromfile, tofile=path, lineterm='')
    patch.diff = '\n'.join(udiff)
    patch.save()


def edit_form(path, branch, repo, user):
    """Supply a page with an edit form."""

    # FIXME: We should be doing find_head directly.
    head = repo.find_head(branch)
    name = repo.getname(path, treename=head)
    if name:
        blob = repo.getblob(name)
        text = blob.contents
    else:
        text = ''

    if text and text[-1] == '\n':
        noeol = "False"
        text = text[0:-1]
    else:
        noeol = True

    form_vars = {'text': text, 'parent': head, 'noeol': noeol}
    page_vars = {'basetemplate': config.basetemplate}

    if user.is_authenticated():
        form_vars['branch'] = branch
        page_vars['submit_label'] = 'Commit'
        page_vars['title'] = 'Edit "%s" from branch "%s"' % (path, branch)
        form = AuthenticatedEditForm(initial=form_vars)
    else:
        page_vars['submit_label'] = 'Submit Patch'
        page_vars['title'] = 'Anonymously Edit "%s"' % (path)
        form = AnonymousEditForm(initial=form_vars)

    page_vars['form'] = form
    return render_to_response('smugform.html', page_vars)


# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Smug Admin URLs.

These URLs are used once for an entire Smug installation.
"""

from views.login import login
from views.logout import logout
from views.merge import merge
from views.page import page
from django.conf.urls.defaults import *


def smugurl(pattern, repo=None):
    """Defines a per-repository URL pattern.

    Works around the brittleness of Django's include() and named urls.
    """
    if not repo:
        raise RuntimeError('Repository must be specified in smugurl.')

    new_pattern = pattern + r'(?P<page>.*)$'
    kwargs = {'session': False, 'repo': repo}
    return (new_pattern, page, kwargs, 'repo-%s' % repo)


urlpatterns = patterns('',
    url(r'^login$', login, name='smug-login',
        kwargs={'session': True, 'template_name': 'login.html'}),
    url(r'^logout$', logout, name='smug-logout',
        kwargs={'session': True, 'template_name': 'logout.html'}),
    url(r'^merge$', merge, name='smug-merge',
        kwargs={'session': True}),
    )

# vim: et sw=4 sts=4



# Gitlib
# Copyright 2008-2009 Andrew McNabb <amcnabb-smug@mcnabbs.org>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Python Git Library

Gitlib is used in Smug, but it's designed to be as general purpose as
possible.  Gitlib uses Git commands for the most part, so it isn't incredibly
fast, but once Git libifies, we should be able to speed it up dramatically.
"""

from repo import Repository, GitError
from objects import Blob, Commit, Tree, TreeEnt

# vim: et sw=4 sts=4



# Gitlib
# Copyright 2008-2009 Andrew McNabb <amcnabb-smug@mcnabbs.org>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Git Repository."""


# TODO:
# - Environment variables for git-commit-tree: GIT_AUTHOR_NAME, etc. (?)
# - Better merging

import os, subprocess
from objects import Blob, Commit, Tree
from index import Index


class GitError(RuntimeError):
    pass


class GitProcError(GitError):
    def __init__(self, msg, gitproc, retcode):
        self.msg = msg
        self.gitproc = gitproc
        self.retcode = retcode

    def __str__(self):
        return self.msg


class GitProc(object):
    def __init__(self, *args, **kwds):
        """Popen variant that executes git with args.

        All of the standard Popen keyword args work.  In particular,
        stdin, stdout, env, and cwd are useful.
        """

        self.args = ['git']
        for arg in args:
            if isinstance(arg, unicode):
                self.args.append(arg.encode('utf-8'))
            else:
                self.args.append(arg)
        if 'env' in kwds:
            env = {}
            for key, value in kwds['env'].iteritems():
                if isinstance(value, unicode):
                    value = value.encode('utf-8')
                env[key] = value
            kwds['env'] = env
        self.gitproc = subprocess.Popen(self.args, **kwds)

    stdin = property(lambda self: self.gitproc.stdin)
    stdout = property(lambda self: self.gitproc.stdout)

    def wait(self):
        """Wait for the process to complete
        
        Return the process' return code.
        """
        self.retcode = self.gitproc.wait()
        return self.retcode

    def read(self, input=None):
        """Read from (and write to) the process and wait for it to complete.

        Returns stdout as a string (so this should not be used to read
        large amounts of data).  Also note that this ignores stderr.
        """
        if isinstance(input, unicode):
            input = input.encode('utf-8')
        stdout, stderr = self.gitproc.communicate(input)
        self.retcode = self.gitproc.returncode
        if self.retcode == 0:
            return stdout
        else:
            command = ' '.join(self.args)
            raise GitProcError('Git command failed: "%s"' % command,
                    self.gitproc, self.retcode)


class Repository(object):
    """Interact with a Git repository.

    The repository may be bare or it may have a working directory.
    """
    def __init__(self, path, create=False):
        self.path = path
        self.shared_repository=None

    def git_exec(self, *args, **kwds):
        """Execute git with args and return a GitProc object.

        Changes working directory to the repository before executing.
        Popen arguments can be specified, such as stdin, stdout, and env.
        """
        kwds['cwd'] = self.path
        gitproc = GitProc(*args, **kwds)
        return gitproc

    def git_command(self, *args, **kwds):
        """Execute git with args and wait for completion.
        
        Save standard output (and return a string on completion) unless
        ignore_output=True is specified.  The environment can be specified
        with the env argument (as a dictionary).
        """
        try:
            ignore_output = kwds['ignore_output']
            del kwds['ignore_output']
        except KeyError:
            ignore_output = False

        if ignore_output:
            kwds['stdout'] = open('/dev/null', 'w')
        else:
            kwds['stdout'] = subprocess.PIPE

        gitproc = self.git_exec(*args, **kwds)
        if ignore_output:
            gitproc.wait()
        else:
            return gitproc.read()

    def create(self):
        """Initializes the Git repository.
        
        Raises a GitError if repository initialization fails.
        """
        try:
            os.makedirs(self.path)
        except OSError, e:
            import errno
            if e.errno != errno.EEXIST:
                raise
        self.git_command('--bare', 'init', ignore_output=True)


    def chmod(self, path):
        """Set the permissions of a path to conform to sharedRepository.
        
        This attempts to mimic the actions of Git's adjust_shared_perm
        function in path.c.
        """
        shared = self.shared_repository
        if shared is None:
            # Find whether it's a shared repository
            try:
                output = self.git_command('config', 'core.sharedRepository')
                shared = output.rstrip()
            except GitError:
                shared = ''
            if shared in ('', 'umask'):
                shared = False
            elif shared in ('1', 'group'):
                shared = 'group'
            elif shared in ('2', 'all', 'world', 'everybody'):
                shared = 'everybody'
            else:
                raise GitError('Unknown value of core.sharedRepository: "%s"'
                        % shared)
            self.shared_repository = shared

        if not shared:
            return

        import stat
        st = os.stat(path)
        mode = st.st_mode

        if (mode & stat.S_IRUSR):
            if (shared == 'group'):
                mode |= stat.S_IRGRP
            elif shared == 'everybody':
                mode |= (stat.S_IRGRP|stat.S_IROTH)

        if (mode & stat.S_IWUSR):
            mode |= stat.S_IWGRP

        if (mode & stat.S_IXUSR):
            if (shared == 'group'):
                mode |= stat.S_IXGRP
            elif shared == 'everybody':
                mode |= (stat.S_IXGRP|stat.S_IXOTH)

        if (stat.S_ISDIR(mode)):
                mode |= stat.S_ISGID

        os.chmod(path, mode)

    # TODO: For writing Tree objects, this should call git-mktree instead of
    # trying to write directly to a file.
    def write(self, gitobj):
        """Write a GitObject (Blob, Tree, Commit or Tag) to the repository.
        """
        dirname = gitobj.name[:2]
        filename = gitobj.name[2:]
        dirpath = os.path.join(self.path, 'objects', dirname)
        filepath = os.path.join(dirpath, filename)

        # If the file already exists, there's nothing to do, since objects
        # with the same object name are identical.
        if not os.path.exists(filepath):
            try:
                os.mkdir(dirpath)
                self.chmod(dirpath)
            except OSError, e:
                import errno
                if e.errno != errno.EEXIST:
                    raise
            f = open(filepath, 'w')
            self.chmod(filepath)
            f.write(gitobj.compressed)
            f.close()

    def find_head(self, branch='master'):
        """Find the tree object for the head of a given branch.

        Return the object name (SHA-1) of the tree object.
        Technically we're returning the id of the commit object, but it
        appears that you can use it anywhere you can use a tree object id.
        """
        try:
            output = self.git_command('show-ref', '--heads', branch)
            return output.split()[0]
        except GitProcError:
            return None

    def save_head(self, commit, oldhead, branch='master'):
        """Save a commit object as to the head ref for the given branch.

        Fail if the value of head that we are replacing is not oldhead.
        Otherwise we would have a nasty race condition.  Set oldhead to
        None if you are creating a new branch.
        """
        branchref = os.path.join('refs/heads', branch)
        if oldhead is None:
            oldhead = ''
        proc = self.git_exec('update-ref', branchref, commit, oldhead,
                stdout=open('/dev/null', 'w'))
        retcode = proc.wait()
        return not retcode

    def getname(self, path, branch=None, treename=None):
        """Lookup the object name (SHA-1) of a path within the specified tree.
        
        The tree can be specified with treename (the object name of a commit
        or tree) or with a branch, in which case the head of the branch will
        be used.  It is illegal to specify both.  The path may have a trailing
        slash, in which case the path must refer to a tree.  However, leading
        slashes are invalid (except in the trivial case of the tree "/", which
        it's really a trailing slash).

        Returns the name of the object at the given path, or None if it cannot
        be found.
        """
        if treename is None:
            branch = 'master'
            treename = self.find_head(branch)
            if treename is None:
                #print "Can't find head for branch=%s" % branch
                return None
        elif branch is not None:
            raise ValueError("Branch and treename cannot both be given to " 
                    "getname()")

        # Mimics the behavior of other tools (e.g., `ls /etc/passwd/` fails)
        if path and path[-1] == '/':
            require_tree = True
            path = path[:-1]
        else:
            require_tree = False

        if not path:
            # Special case: `git ls-tree ""` does not show the name of the
            # root tree.
            commit = self.getcommit(treename)
            name = commit.tree
        else:
            if path[0] == '/':
                raise GitError("getname() was given a path with a leading /.")

            ls = self.git_command('ls-tree', treename, path)
            name = None
            if ls:
                fields = ls.split()
                objtype = fields[1]
                if not require_tree or objtype == 'tree':
                    name = fields[2]

        return name

    def gettype(self, name):
        """Get the type of the object identified by the given object name.
        
        For convenience, None is returned if name is not specified.
        """
        if name:
            output = self.git_command('cat-file', '-t', name)
            return output.split()[0]

    def getblob(self, name):
        """Retrieves the blob with the given object name as a Blob object.
        
        For convenience, None is returned if name is not specified.
        """
        if name:
            contents = self.git_command('cat-file', 'blob', name)
            return Blob(contents)

    def getcommit(self, name):
        """Retrieves the blob with the given object name as a Blob object."""
        contents = self.git_command('cat-file', 'commit', name)
        return Commit(contents)

    def gettree(self, name):
        """Retrieves the tree with the given object name as a Tree object."""
        contents = self.git_command('cat-file', 'tree', name)
        return Tree(contents)

    def walktree(self, name, reverse=False):
        """Recursively walks through the tree with the given object name.
        
        Yields (path, object_name) pairs which are sorted within each subtree.
        If reverse is True, entries are sorted in reverse order.
        """
        tree = self.gettree(name)
        for path, entry in tree.entries(reverse):
            if entry.is_dir():
                for subpath, name in self.walktree(entry.name, reverse):
                    yield '/'.join((path, subpath)), name
            else:
                yield path, entry.name

    def unpack_file(self, name):
        """Open the blob with SHA-1 name, and save it to a temporary file.

        Return the filename of the temporary file, which the caller is
        responsible for deleting.
        """
        output = self.git_command('unpack-file', name)
        return output.rstrip()

    def readtree(self, tree):
        """Read a tree into an Index."""
        return Index(self, tree)

    # TODO: Just write a Commit object and call git-update-ref instead of
    # calling the commit function.
    def commit(self, tree, parent_ids, changelog, author_name=None,
            author_email=None):
        """Create a new commit object for the given tree.

        Specify the changelog as a string and the parent commits as list
        of ids of tree objects.  The author name and email may optionally
        be specified.  Return the id of the new commit.
        """
        env = {}
        if author_name:
            env['GIT_AUTHOR_NAME'] = author_name
        if author_email:
            env['GIT_AUTHOR_EMAIL'] = author_email

        args = ['commit-tree', tree]
        for name in parent_ids:
            args += ['-p', name]
        kwds = {'stdin': subprocess.PIPE, 'stdout': subprocess.PIPE,
                'env': env}
        gitproc = self.git_exec(*args, **kwds)
        output = gitproc.read(changelog)

        return output.rstrip()

    def commit_add(self, path, text, parent, changelog, author_name=None,
            author_email=None):
        """Add or change a single file in the tree and commit.
        
        Return the object name (SHA-1) of the new git object.
        """
        if parent:
            index = self.readtree(parent)
        else:
            index = Index(self, None)
        blob = Blob()
        blob.text = text
        blob.freeze()
        self.write(blob)
        index.add(path, blob.name)
        newtree = index.write()
        parent_ids = []
        if parent:
            parent_ids.append(parent)
        name = self.commit(newtree, parent_ids, changelog, author_name,
                author_email)
        return name

    def find_mergebase(self, branch1, branch2):
        """Finds the best common ancestor between the heads of two branches.
        
        Returns (current, mergebase, other), the object names of three commits.
        """
        current = self.find_head(branch1)
        other = self.find_head(branch2)
        output = self.git_command('merge-base', current, other)
        mergebase = output.rstrip()
        return (current, mergebase, other)

    # TODO: in merge, set author_name and author_email on commit
    def merge(self, branch, current, mergebase, other_commit,
            resolutions=None):
        """Merges in changes from other_commit using 3-way merge.

        Automatically resolves any trivial changes.  Returns a list of
        conflicts.

        Don't call this if merging is unnecessary or if a fast-forward merge
        would be most appropriate.  If there are any conflicts, you will need
        to come back and tell us what to do.

        Specify resolutions if you have already figured out how to resolve the
        conflicts that will arise.  This should be a dictionary mapping paths
        to contents.  We naively assume that file permissions should always
        be 100644.  Someday that may need to change.
        
        We will return a dict of any other unresolved conflicts:

        conflicts[path] = 'Text of file with conflict markers.'

        Note:
            mergebase = self.find_mergebase(self.commit, other_commit)
        """
        gitlib_res = {}
        MODE = '100644'
        for path, contents in resolutions.items():
            blob = gitlib.Blob(contents)
            self.write(blob)
            gitlib_res[path] = (blob.name, MODE)

        # IN PROGRESS:
        index = self.readtree(current)
        conflicts = index.merge(mergebase, other_commit, gitlib_res)
        if not conflicts:
            commitmsg = 'Merge made by Gitlib'
            if resolutions:
                commitmsg += ' with manual conflict resolution.'
            else:
                commitmsg += ' automatically.'

            newtree = index.write()
            parents = [current, other_commit]
            newcommit = self.commit(newtree, parents, commitmsg)
            self.save_head(newcommit, current, branch)
            return {}
        else:
            # Get rid of the "[" and "]" when we stop supporting Python 2.3:
            return dict([(c.path, c.merge_file()) for c in conflicts])


# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from django import http
from django.conf import settings
from smug.load import get_content

def show(request, page, repo, branch, raw=False, **kwds):
    """Render a page straight from Git."""

    filtered_file = get_content(page, branch, repo, raw=raw, request=request)

    if filtered_file.is_directory():
        if page and not page.endswith('/'):
            return http.HttpResponseRedirect(page + '/')
        else:
            # TODO: add a configuration option for generated directory
            # indexes (in which case this would not be a 404).
            raise http.Http404

    if filtered_file.mimetype.startswith('text/'):
        content_type = '%s; charset=%s' % (filtered_file.mimetype,
                settings.DEFAULT_CHARSET)
    else:
        content_type = filtered_file.mimetype
    response = http.HttpResponse(content_type=content_type)
    # note: Python 2.4 has no rpartition
    #dirname, slash, filename = page.rpartition('/')
    split = page.rsplit('/', 1)
    if len(split) == 2:
        dirname, filename = split
    else:
        dirname = filename = ''
    response['Content-Disposition'] = 'inline; filename=%s' % filename
    response.write(filtered_file.content)
    return response


# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Login form.

Note that this is different from Django's login form because we need to be
do SSL session redirecting, and Django's form breaks it.
"""

from django.http import HttpResponseRedirect, HttpResponseBadRequest
from django.shortcuts import render_to_response
from django.contrib import auth
from django.contrib.auth import forms, REDIRECT_FIELD_NAME

LOGIN_TEMPLATE = 'login.html'

# TODO: make it so the name doesn't conflict with auth.login
def login(request, **kwds):
    redirect = request.REQUEST.get(REDIRECT_FIELD_NAME, '')

    if request.POST:
        f = forms.AuthenticationForm(request=request, data=request.POST)
        if f.is_valid():
            # check from django auth: "make sure redirect isn't garbage:"
            if not redirect or '//' in redirect or ' ' in redirect:
                return HttpResponseBadRequest('Invalid redirect.')
            auth.login(request, f.get_user())
            if request.session.test_cookie_worked():
                request.session.delete_test_cookie()
            return HttpResponseRedirect(redirect)
    else:
        f = forms.AuthenticationForm()

    request.session.set_test_cookie()

    context = {'form': f, REDIRECT_FIELD_NAME: redirect}
    return render_to_response(LOGIN_TEMPLATE, context)

# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from django.db import models

STATUS_CHOICES = (
        ('s', 'Submitted'),
        ('a', 'Approved'),
        ('r', 'Rejected'),
        ('w', 'Withdrawn'),
        )

class Patch(models.Model):
    status = models.CharField('Status', max_length=1, choices=STATUS_CHOICES,
            default='i')
    timestamp = models.DateTimeField(auto_now=True)
    path = models.CharField('Path', max_length=200)
    diff = models.TextField('Diff')
    changelog = models.CharField('Changelog', max_length=80)
    parent = models.CharField('Parent Commit', max_length=40)

    class Meta:
        verbose_name_plural="Patches"

# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from smug.pathops.show import show


def raw(request, page, repo, branch, **kwds):
    """Render a page straight from Git without any filtering.

    Give the correct mime type (according to file extension), regardless of
    what the filterconf.py file says to do.
    """
    return show(request, page, repo, branch, raw=True)


# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Template helpers for Django."""

BRANCH = 'master'
PROTOCOL = 'smug://'

from django.template import TemplateDoesNotExist
from django.conf import settings
#from django.http import HttpResponsePermanentRedirect
from django.http import HttpResponseRedirect
import config

def smug_load_template_source(template_name, template_dirs=None):
    """Extract a template from the Git repository"""

    import os
    if template_dirs is None:
        template_dirs = settings.TEMPLATE_DIRS

    for template_dir in template_dirs:
        if template_dir.startswith(PROTOCOL):
            repokey, subdir = template_dir[len(PROTOCOL):].split('/', 1)
            repo = config.repositories[repokey]

            path = '/'.join((subdir, template_name))
            name = repo.getname(path, BRANCH)
            blob = repo.getblob(name)
            if blob is not None:
                full_url = '/'.join((template_dir, path))
                return blob.contents, full_url
    raise TemplateDoesNotExist(template_name)

smug_load_template_source.is_usable = True


try:
    http_base = settings.HTTP_BASEURL
except AttributeError:
    http_base = 'http://'

try:
    https_base = settings.HTTPS_BASEURL
except AttributeError:
    https_base = 'https://'

def ssl_redirect(request, ssl=True):
    """Redirect to https if ssl=True or to http if ssl=False.

    The values configured in settings.py as HTTP_BASEURL and HTTPS_BASEURL
    are required if they differ in any way other than "http" vs. "https".
    """
    if (ssl is None) or (ssl == request.is_secure()):
        return None

    old_uri = request.build_absolute_uri()

    if ssl and old_uri.startswith(http_base):
        # Redirect to https
        new_uri = https_base + old_uri[len(http_base):]
    elif (not ssl) and old_uri.startswith(https_base):
        # Redirect to http
        new_uri = http_base + old_uri[len(https_base):]
    else:
        raise RuntimeError("Base URLs are inconsistent in ssl_redirect()")

    #return HttpResponsePermanentRedirect(new_uri)
    return HttpResponseRedirect(new_uri)


class SecureSessionMiddleware(object):
    """Redirect paths between http and https as requested by urlpatterns.
    
    Check for 'secure' in a view's kwargs.  If not set or None, do nothing.
    If True, redirect to https if secure sessions are requested.  If False,
    redirect to http.

    The values configured in settings.py as HTTP_BASEURL and HTTPS_BASEURL
    are required if they differ in any way other than "http" vs. "https".
    """
    def process_view(self, request, view_func, view_args, view_kwargs):
        try:
            ssl = view_kwargs['session']
            del view_kwargs['session']
        except KeyError:
            ssl = None

        if settings.SESSION_COOKIE_SECURE:
            return ssl_redirect(request, ssl)
        else:
            return None

# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO: allow filters to give errors in a meaningful way.  The best way to do
# this is probably to add HTTP return codes to filters (and cache error
# messages).  (maybe make it timeout at 5 minutes instead 30 minutes)

"""Smug content filters.

A filter is a function that takes a `file` and optional keyword args.  The
file is initialized from the source file, so a no-op filter can simply return.
The attributes of the file that are usually changed are `content` and
`mimetype`.

By the way, we don't feel bad shadowing the builtin `file` because it's going
away in Python 3.0 anyway.
"""

from smug import config

# TODO: allow filters to be chained (when filter names separated by commas).
def get_filter(using):
    """Return the filter function for `using`."""
    return FILTERS[using]


def plaintext(file, **kwds):
    """Retrieve content as-is but with mimetype text/plain."""
    
    file.mimetype = 'text/plain'


def extract_title(s):
    """Extract the title from a string.

    The title must be contained in the first line of the string and must be of
    the form:

    <!-- title: Title Goes Here -->
    """
    OPEN = "<!--"
    CLOSE = "-->"
    LABEL = "title"
    try:
        newline = s.index('\n')
        start = s.index(OPEN, 0, newline) + len(OPEN)
        end = s.index(CLOSE, start, newline)
    except ValueError:
        return ''

    comment = s[start:end]
    key, value = comment.split(':', 1)

    if key.strip().lower() != LABEL:
        return ''

    return value.strip()


def present(file, repo, **kwds):
    """Render a Django template with the content as a context variable."""

    try:
        template = kwds['template']
    except KeyError:
        template = config.basetemplate
    request = kwds['request']

    raw_content = file.content
    if not raw_content:
        return

    from django.core.urlresolvers import reverse
    baseurl = reverse('repo-%s' % repo.name, kwargs={'page': ''})
    context_dict = {'baseurl': baseurl}

    # Setup the menu of Smug actions
    from menu import SmugMenu
    menu = SmugMenu(file.path, request)
    menu.add_query('Edit', 'edit', True)
    context_dict['smugmenu'] = menu

    from django.utils.safestring import mark_safe

    context_dict['title'] = extract_title(raw_content)
    context_dict['content'] = mark_safe(raw_content)

    from django.template.loader import render_to_string
    file.content = render_to_string(template, context_dict)


def rst2html(file, repo, **kwds):
    """Render a Django template with the content as a context variable."""

    template = kwds.get('template', config.basetemplate)
    request = kwds['request']

    raw_content = file.content
    if not raw_content:
        return

    from django.core.urlresolvers import reverse
    baseurl = reverse('repo-%s' % repo.name, kwargs={'page': ''})
    context_dict = {'baseurl': baseurl}

    # Setup the menu of Smug actions
    from menu import SmugMenu
    menu = SmugMenu(file.path, request)
    menu.add_query('Edit', 'edit', True)
    context_dict['smugmenu'] = menu

    # Get HTML parts from reST
    from docutils import core
    parts = core.publish_parts(source=raw_content,writer_name='html')

    from django.utils.safestring import mark_safe
    
    context_dict['title'] = parts['title']
    context_dict['content'] = mark_safe(parts['html_body'])

    from django.template.loader import render_to_string
    file.content = render_to_string(template, context_dict)
    file.mimetype = 'text/html'


# TODO: figure out what images the latex source needs, and make them available
# to pdflatex.
def latex2pdf(file, **kwds):
    """Render a PDF from a tex template and send it to the user."""
    import subprocess, shutil, tempfile

    # TODO: Read PDFLATEX_BIN from the settings file or something.
    PDFLATEX_BIN = 'pdflatex'

    tmpdir = tempfile.mkdtemp('smugtex')
    texfile = tmpdir + '/djangotex.tex'
    pdffile = tmpdir + '/djangotex.pdf'

    # Alternatively, we could pipe the contents to LaTeX both times, but this
    # makes for easier debugging and might even be faster.
    open(texfile, 'w').write(file.content)

    # NOTE: LaTeX often needs to be run multiple times.  Here we assume it
    # needs exactly two runs.

    # pdflatex outputs 0 for good, 1 for so-so, and 255 for really bad
    args = (PDFLATEX_BIN, '-draftmode', '-interaction', 'batchmode', texfile)
    retcode = subprocess.call(args, cwd=tmpdir)
    if retcode != 0 and retcode != 1:
        #raise PDFLatexFailed
        return

    args = (PDFLATEX_BIN, '-interaction', 'batchmode', texfile)
    retcode = subprocess.call(args, cwd=tmpdir)
    if retcode != 0 and retcode != 1:
        #raise PDFLatexFailed
        return

    file.content = open(pdffile, 'r').read()
    file.mimetype = 'application/pdf'

    # TODO: make a config option that determines whether or not to do the
    # rmtree (sometimes you need the files for debugging):
    shutil.rmtree(tmpdir)


def rst2latex(file, **kwds):
    """Render a ReST source file as a latex source document."""

    raw_content = file.content
    if not raw_content:
        return

    file.mimetype = "application/x-latex"

    # Get latex from reST
    from docutils import core
    file.content = core.publish_string(source=raw_content, writer_name='latex')


# TODO: Do proper nested filters.
def rst2latex2pdf(file, **kwds):
    """Render a ReST source file as a latex-generated pdf document."""

    # First, render the latex source file.
    rst2latex(file, **kwds)
    # Then render the pdf.
    latex2pdf(file, **kwds)


def plot2png(file, **kwds):
    """Render a gnuplot script as a png image."""

    from subprocess import Popen, PIPE

    REMOVE = 'set terminal'
    INSERT = 'set terminal png'

    original = file.content.splitlines()
    script = [INSERT]
    script += [line for line in original if not line.startswith(REMOVE)]

    proc = Popen(('gnuplot',), stdin=PIPE, stdout=PIPE)
    proc.stdin.write('\n'.join(script))
    proc.stdin.close()
    proc.wait()
    file.content = proc.stdout.read()
    file.mimetype = "image/png"


FILTERS = dict([(f.__name__, f) for f in (plaintext, present, rst2html,
        rst2latex, rst2latex2pdf, plot2png)])

# vim: et sw=4 sts=4



#!/usr/bin/env python

import doctest
import gitlib


doctest.testmod(gitlib.objects)
doctest.testmod(gitlib.repo)
doctest.testmod(gitlib.index)
doctest.testfile('../docs/gitlib.txt')


# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

DEFAULT_BRANCH = 'master'

from django.http import HttpResponseBadRequest
from smug import config
from smug.pathops.edit import edit
from smug.pathops.raw import raw
from smug.pathops.show import show

PATHOPS = {'edit': edit, 'raw': raw}

def page(request, page=None, repo=None):
    if repo is None:
        raise RuntimeError('URLs configuration error.  No repo specified.')
    repo = config.repositories[repo]

    if '//' in page:
        error = 'URLs must not contain double slashes ("//").'
        return HttpResponseBadRequest(error)

    branch = request.GET.get('branch', DEFAULT_BRANCH)
    pathop = None
    for name, func in PATHOPS.iteritems():
        if name in request.GET:
            if pathop:
                error = 'Only one pathop may be specified.'
                return HttpResponseBadRequest(error)
            else:
                pathop = func

    if pathop is None:
        pathop = show

    return pathop(request, page, repo, branch)


# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from django.http import HttpResponseRedirect
from django.shortcuts import render_to_response
from django.contrib import auth

LOGOUT_TEMPLATE = 'logout.html'

def logout(request, **kwds):
    auth.logout(request)

    redirect = request.REQUEST.get(auth.REDIRECT_FIELD_NAME, '')
    if redirect:
        return HttpResponseRedirect(redirect)
    else:
        return render_to_response(LOGOUT_TEMPLATE, {})


# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

DEFAULT_TEMPLATE = 'template.html'
DEFAULT_CACHE_TIME = 30 * 60
DEFAULT_HEAD_TIMEOUT = 20
# An alternative DEFAULT_MIMETYPE would be application/octet-stream
DEFAULT_MIMETYPE = 'text/plain'
DEFAULT_ADMIN_URL = '/admin/'

from django.conf import settings
import gitlib

def _get_setting(name, default):
    try:
        return getattr(settings, name)
    except AttributeError:
        return default

# Setup revsys
repositories = {}
try:
    SMUG_REPOSITORIES = settings.SMUG_REPOSITORIES
except AttributeError:
    SMUG_REPOSITORIES = None
if SMUG_REPOSITORIES:
    for name, path in SMUG_REPOSITORIES.iteritems():
        repo = gitlib.Repository(path)
        repositories[name] = repo
        repo.name = name
    del repo

basetemplate = _get_setting('SMUG_TEMPLATE', DEFAULT_TEMPLATE)

# How long to cache an ordinary page.
cache_time = _get_setting('SMUG_CACHE_TIME', DEFAULT_CACHE_TIME)
# How long to cache the head ref.
head_timeout = _get_setting('SMUG_HEAD_TIMEOUT', DEFAULT_HEAD_TIMEOUT)
# Admin directory (note, None means admin not installed)
admin_url = _get_setting('ADMIN_URL', DEFAULT_ADMIN_URL)

# vim: et sw=4 sts=4



# Gitlib
# Copyright 2008-2009 Andrew McNabb <amcnabb-smug@mcnabbs.org>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Git Object implementations.

Real git objects are immutable, but we need to provide a mutable interface.
As a result, a GitObject starts off as "fresh" (mutable).  Once it is
"frozen", the object name and other properties can be computed, but nothing
can be changed.  To make a change, a new GitObject must be created using the
"copy" method.
"""


import codecs
hex_codec = codecs.lookup('hex_codec')


def frozen_only_property(cache_name):
    """A decorator for defining properties that require freezing.

    Usage is pretty simple.  Just give it the name of the cache variable to
    use.  The name of your function will be the name of the property.

    >>> class SomeGitObject(GitObject):
    ...     _type = 'fake_type'
    ...     hello = None
    ...     @frozen_only_property('_cache_variable')
    ...     def xyz(self):
    ...         return self.hello
    >>> test = SomeGitObject()
    >>> test.hello = 4
    >>> test.xyz
    Traceback (most recent call last):
        ...
    AttributeError: GitObject is not frozen.
    >>> test.freeze()
    >>> test.xyz
    4
    >>>

    Note that after freezing, the value can't be changed:
    >>> test.hello = 5
    >>> test.xyz
    4
    >>>
    """
    def decorator(function):
        """Takes a function which generates the read-only property."""
        def getter(self):
            if not self.frozen:
                raise AttributeError("GitObject is not frozen.")
            try:
                value = getattr(self, cache_name)
            except AttributeError:
                value = None
            if value is None:
                value = function(self)
                setattr(self, cache_name, value)
            return value

        return property(getter, doc=function.__doc__)

    return decorator


def fresh_only_property(cache_name):
    """A property that can only be set before freezing.

    >>> class SomeGitObject(GitObject):
    ...     _type = 'fake_type'
    ...     xyz = fresh_only_property('_cache_variable')
    >>> test = SomeGitObject()
    >>> test.xyz = 4
    >>> test.freeze()
    >>> test.xyz = 5
    Traceback (most recent call last):
        ...
    AttributeError: Attribute can only be set before freezing.
    >>> test.xyz
    4
    >>>
    """
    def getter(self):
        try:
            return getattr(self, cache_name)
        except AttributeError(self):
            setattr(self, cache_name, None)
            return None

    def setter(self, value):
        if self.frozen:
            raise AttributeError('Attribute can only be set before freezing.')
        setattr(self, cache_name, value)

    return property(getter, setter)


class GitObject(object):
    """A Git object, which may be a blob, a tree, a commit, or a tag.

    Once a Git object is frozen, it cannot be changed.  However, it can be
    copied.  Certain attributes, such as `gitobject` and `compressed` are not
    available until the object is frozen.  A Git object can be initialized in
    a frozen form by calling the constructor with the raw contents.  Note that
    the contents and fields are always bytes, not Unicode.

    Inheriting objects should set copy, parse_contents, and make_contents.
    """
    _type = None

    _frozen = False
    frozen = property(lambda self: self._frozen,
            doc="A frozen GitObject cannot be modified")

    def __init__(self, contents=None):
        if not self._type:
            raise TypeError("GitObject cannot be instantiated directly.")

        if contents is not None:
            self.set_contents(contents)
            self.freeze()

    def freeze(self):
        """Freeze the Git object"""
        self._frozen = True

    def __str__(self):
        return self.contents

    def set_contents(self, contents):
        """Set contents of the GitObject, not including headers"""
        if self.frozen:
            raise AttributeError("Can't modify a frozen GitObject")
        else:
            self.parse_contents(contents)

    @frozen_only_property('_contents')
    def contents(self):
        """Contents of the GitObject, not including headers."""
        return self.make_contents()

    @frozen_only_property('_gitobject')
    def gitobject(self):
        """Uncompressed GitObject."""
        size = len(self.contents)
        return '%s %s\0%s' % (self._type, str(size), self.contents)

    @frozen_only_property('_compressed')
    def compressed(self):
        """Compressed GitObject."""
        import zlib
        return zlib.compress(self.gitobject)

    @frozen_only_property('_name')
    def name(self):
        """Object name (SHA1 hash of GitObject)."""
        try:
            import hashlib
            hash_constructor = hashlib.sha1
        except ImportError:
            import sha
            hash_constructor = sha.sha
        hash_obj = hash_constructor(self.gitobject)
        return hash_obj.hexdigest()

    def copy(self, contents):
        """Creates a mutable copy of a Git object.

        This must be implemented by individual GitObject classes.
        """
        raise NotImplementedError

    def make_contents(self):
        """Consolidates individual fields to create the full Git contents.

        This must be implemented by individual GitObject classes.
        """
        raise NotImplementedError

    def parse_contents(self, contents):
        """Parses fields from the contents of a Git object.

        This must be implemented by individual GitObject classes.
        """
        raise NotImplementedError


class Blob(GitObject):
    """A blob object that may be added to a Git repository.

    The text of a blob is bytes (binary files may not be valid UTF-8).
    """

    _type = 'blob'

    _text = ''
    text = fresh_only_property('_text')

    def copy(self):
        blob = Blob()
        blob._text = self._text
        return blob

    def make_contents(self):
        return self._text

    def parse_contents(self, contents):
        self._text = contents


class TreeEnt(object):
    r"""Entry in a Tree.
    
    All attributes are read-only.

    >>> ent = TreeEnt('abc', 'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391',
    ...         '100644')
    >>> expected = ('100644 abc\x00' +
    ...     '\xe6\x9d\xe2\x9b\xb2\xd1\xd6CK\x8b)\xaewZ\xd8\xc2\xe4\x8cS\x91')
    >>> str(ent) == expected
    True
    >>>
    """
    filename = property(lambda self: self._filename)
    name = property(lambda self: self._name)
    mode = property(lambda self: self._mode)

    def __init__(self, filename, name, mode):
        self._filename = filename
        self._name = name
        self._mode = mode

    def is_dir(self):
        import stat
        mode = int(self.mode, 8)
        return stat.S_ISDIR(mode)

    def __str__(self):
        raw_name, bytes_consumed = hex_codec.decode(self._name)
        return ''.join((self._mode, ' ', self._filename, '\0', raw_name))


class Tree(GitObject):
    """A tree object that may be added to a Git repository.
    
    >>> t = Tree()
    >>> t.add_file('abc', 'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391', 'blob')
    True
    >>> t.add_file('xyz', '484ba93ef5b0aed5b72af8f4e9dc4cfd10ef1a81', 'blob')
    True
    >>> contents = t.make_contents()
    >>>

    >>> u = Tree()
    >>> u.parse_contents(contents)
    >>>
    """
    _type = 'tree'

    def __init__(self, *args):
        self._filemap = {}
        GitObject.__init__(self, *args)

    def make_contents(self):
        return ''.join([str(self._filemap[k]) for k in sorted(self._filemap)])

    def parse_contents(self, contents):
        import sys
        while contents:
            # Mode.
            space = contents.index(' ')
            mode = contents[:space]

            # Filename.
            null = contents.index('\0', space + 1)
            filename = contents[space + 1: null]

            # Object Name (SHA-1).
            name_length = 20
            start_of_name = null + 1
            end_of_name = start_of_name + name_length
            raw_name = contents[start_of_name: end_of_name]
            name, bytes_consumed = hex_codec.encode(raw_name)

            # Create the Tree Entry.
            entry = TreeEnt(filename, name, mode)
            self._filemap[filename] = entry

            contents = contents[end_of_name:]

    def add_file(self, filename, name, type, mode=None, replace=False):
        """Adds an object to the tree.

        The filename, name (SHA-1), and type must be specified.  The mode
        defaults to '100644' for blobs and '040000' for trees.  If replace
        is given and True, then any existing file with the same filename
        will be replaced.

        Returns True if the operation succeeded, and False otherwise (i.e.,
        filename already exists).  An exception is raised if the Tree is
        frozen.
        """
        assert not self.frozen

        if (not replace) and (filename in self._filemap):
            return False
        if '/' in filename:
            raise ValueError('Filenames may not contain slashes.')

        if mode is None:
            if type == 'blob':
                mode = '100644'
            elif type == 'tree':
                mode = '040000'

        self._filemap[filename] = TreeEnt(filename, name, mode)
        return True

    def entries(self, reverse=False):
        """Iterates over the tree entries in sorted order.
        
        Yields (path, entry) pairs.
        """
        return sorted(self._filemap.iteritems(), reverse=reverse)


class Commit(GitObject):
    r"""A commit object that may be added to a Git repository.

    >>> tree = '7c6cf9b169ab51858fbbb77f352d0de249bb1e97'
    >>> parent = 'ecf5e79f6dc40b9f517baf6de5755e31ed4188fb'
    >>> author = 'Andrew McNabb <amcnabb@mcnabbs.org> 1219616814 -0600'
    >>> committer = author
    >>> message = 'added mdadm.conf\n'
    >>> contents = ('tree %s\nparent %s\nauthor %s\ncommitter %s\n\n%s' %
    ...     (tree, parent, author, committer, message))
    >>> commit = Commit(contents)
    >>> commit.freeze()
    >>> commit.contents == contents
    True
    >>>
    """

    _type = 'commit'

    _tree = ''
    tree = fresh_only_property('_tree')
    _parents = ()
    parents = fresh_only_property('_parents')
    _author = ''
    author = fresh_only_property('_author')
    _committer = ''
    committer = fresh_only_property('_committer')
    _message = ''
    message = fresh_only_property('_message')

    def copy(self):
        commit = Commit()
        commit.tree = self.tree
        commit.parents = list(self.parents)
        commit.author = self.author
        commit.committer = self.committer
        commit.message = self.message
        return commit

    def make_contents(self):
        lst = []
        lst.append('tree %s' % self.tree)
        for parent in self.parents:
            lst.append('parent %s' % parent)
        lst.append('author %s' % self.author)
        lst.append('committer %s' % self.committer)
        lst.append('')
        lst.append(self.message)
        return '\n'.join(lst)

    def parse_contents(self, contents):
        """Sets the individual fields of a commit based on its contents."""
        header, message = contents.split('\n\n', 1)
        for line in header.splitlines():
            key, value = line.split(None, 1)
            if key in ('tree', 'author', 'committer'):
                setattr(self, key, value)
            elif key == 'parent':
                self.parents = self.parents + (value,)
            else:
                print 'Warning: unknown key "%s"' % key
        self.message = message

        # Sanity check:
        #assert self.make_contents() == contents
        """
        if self.make_contents() != contents:
            print "Error! Contents doesn't match str(self)!"
            print str(self)
            print contents
            raise RuntimeError
        """


# vim: et sw=4 sts=4



# See http://peak.telecommunity.com/DevCenter/setuptools for setuptools info.

import ez_setup
ez_setup.use_setuptools()
from setuptools import setup, find_packages
from smug import VERSION

long_description = """Smug is a live-editing content system backed by a Git repository. It behaves like a wiki, except that the content is fully backed by a revision control system.
"""

setup(
    name='smug',
    version=VERSION,
    packages=find_packages(),
    include_package_data=True,
    requires=['django'],
    provides=['smug', 'gitlib'],
    author='Andrew McNabb',
    author_email='amcnabb@mcnabbs.org',
    url='http://www.mcnabbs.org/andrew/smug/',
    description='smugly superior to wikis',
    long_description=long_description,
    license='GNU GPLv3 or later',
    platforms=['linux', 'Apple OS X'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.4',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Version Control',
        'Topic :: Text Processing :: Markup',
        ]
    )



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core.urlresolvers import reverse
from itertools import chain
import config

class SmugMenu(object):
    """Manage a menu of Smug actions.
    
    The `page` attribute holds the current request's path within the
    repository, while the `path` attribute holds the request's path on the
    site.
    """
    def __init__(self, page, request):
        self.items = []
        self.lastitems = []
        self.page = page
        self.path = request.path

        self.authenticated = request.user.is_authenticated()

        if self.authenticated:
            # place the logout button
            # TODO: make this configurable
            # note: if you leave off the redirect, it will do a logout page
            logout_url = reverse('smug-logout')
            url = '%s?%s=%s' % (logout_url, REDIRECT_FIELD_NAME, self.path)
            self.lastitems.append(('Logout(%s)' % (request.user.username), url))
            # place the admin button
            if config.admin_url is not None:
                self.lastitems.append(('Admin', config.admin_url))
        else:
            login_url = reverse('smug-login')
            url = '%s?%s=%s' % (login_url, REDIRECT_FIELD_NAME, self.path)
            self.lastitems.append(('Login', url))

    def add_abs(self, name, url, auth=False):
        """Add a Smug action with absolute url.
        
        Specify a name, url, and whether authentication is required.
        """
        if not auth or self.authenticated:
            self.items.append((name, url))

    def add_query(self, name, query, auth=False):
        """Add a Smug action that adds a query to the current page.

        Specify a name, query, and whether authentication is required.  The
        query is of the form "key1=value1&key2=value2".
        """
        if not auth or self.authenticated:
            self.items.append((name, '?'.join((self.path, query))))

    def __iter__(self):
        return chain(self.items, self.lastitems)

# vim: et sw=4 sts=4



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import admin
from models import Patch

class PatchAdmin(admin.ModelAdmin):
    radio_fields = {'status': admin.VERTICAL}

admin.site.register(Patch, PatchAdmin)



# Smug
# Copyright 2008 Andrew McNabb <amcnabb-smug@mcnabbs.org>
#
# This file is part of Smug.
#
# Smug is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Smug is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.shortcuts import render_to_response
from django import forms

from smug.config import basetemplate, repositories


class MergePrepForm(forms.Form):
    """Form for first part of merge."""
    branch1 = forms.CharField(label='Branch to Merge Into')
    branch2 = forms.CharField(label='Branch to Merge From')
    do_merge = forms.CharField(widget=forms.widgets.HiddenInput)

class ConflictForm(forms.Form):
    """Form for first part of merge."""
    branch1 = forms.CharField(widget=forms.widgets.HiddenInput)
    branch2 = forms.CharField(widget=forms.widgets.HiddenInput)
    current = forms.CharField(widget=forms.widgets.HiddenInput)
    mergebase = forms.CharField(widget=forms.widgets.HiddenInput)
    other = forms.CharField(widget=forms.widgets.HiddenInput)
    do_merge = forms.CharField(widget=forms.widgets.HiddenInput)

def merge(request, repo=None, **kwds):
    """Attempt to merge any missing commits from one branch into another.

    URL looks like:
        http://smug/!merge?from=branch2&into=branch1
    Normal users will usually do:
        http://smug/!merge?from=master&into=username/mybranch
    in order to bring mybranch up-to-date.  Editors will usually do:
        http://smug/!merge?from=someuser/somebranch&into=master
    in order to incorporate a user's changes into the mainstream.

    We will attempt to do a a trivial merge.  If there are conflicts, we
    will redirect the user to a conflict resolution page, the details of
    which are not yet known.
    """
    if repo is None:
        raise RuntimeError('URLs configuration error.  No repo specified.')
    repo = repositories[repo]

    branch1 = request.GET.get('branch1', '')
    branch2 = request.GET.get('branch2', '')
    # If branches are specified in POST, they override those from GET:
    branch1 = request.POST.get('branch1', branch1)
    branch2 = request.POST.get('branch2', branch2)

    current = request.POST.get('current', '')
    other = request.POST.get('other', '')
    mergebase = request.POST.get('mergebase', '')

    # See if the MergePrepForm has already been filled out:
    try:
        request.POST['do_merge']
        do_merge = True
    except KeyError:
        do_merge = False

    if not branch1 or not branch2:
        do_merge = False
    # TODO: Make sure that both branches exist, if given.

    if do_merge:
        if not current or not mergebase or not other:
            current, mergebase, other = repo.prep_merge(branch1, branch2)

        if mergebase == other:
            return HttpResponse("Nothing to do (we're newer than the other branch.")

        # It turns out that 3-way merge is a slow way of doing a fast-forward
        # merge in this case.  We should still probably implement the
        # fast-forward merge, though, because that way the merge wouldn't add
        # an extra commit message into the logs.
        #elif mergebase == current:
            #return HttpResponse('We need to do a fast-forward merge.')

        # See if we can do the merge as-is:
        resolutions = {}
        for key, contents in request.POST.items():
            if key.startswith('conflict_'):
                path = key.replace('conflict_', '', 1)
                resolutions[path] = contents
        new_conflicts = repo.merge(branch1, current, mergebase, other,
                resolutions)

        if new_conflicts:
            return HttpResponse('Conflicts form goes here.')
        else:
            return HttpResponse('Merge successful: branch1=%s.' % branch1)
        # End with an HttpResponseRedirect
    else:
        # Present the user with an explanation of what is about to happen
        # and give them a "Merge" button.

        # TODO: Maybe we should show a picture of how the branches relate a la
        # gitk on this page.

        form_vars = {'do_merge': 'do_merge',
                'branch1': branch1, 'branch2': branch2}
        form = MergePrepForm(initial=form_vars)

        title = 'Merge from branch "%s" into branch "%s"' % (branch2, branch1)
        page_vars = {'title': title, 'form': form,
                'submit_label': 'Merge',
                'basetemplate': basetemplate}
        return render_to_response('smugform.html', page_vars)

# vim: et sw=4 sts=4



# Gitlib
# Copyright 2008-2009 Andrew McNabb <amcnabb-smug@mcnabbs.org>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Git Index (for modifying and merging trees)."""


import os
from objects import Blob


# TODO: Note that we could create a Tree object and write it with git-mktree
# instead of having to create an entire Index.  I would expect this to reduce
# overhead for any case without merging.
class Index(object):
    """Git index or directory cache

    All commits and merges are staged in an index file.  The Index class
    manages such a file.  Specify the commit to read from as a SHA-1.
    """

    def __init__(self, repository, commit, indexfile=None):
        self.repository = repository
        self.commit = commit
        if indexfile is None:
            from tempfile import mkstemp
            fd, indexfile = mkstemp(prefix='python-gitlib-index')
            os.close(fd)
            self.tempfile = True
        else:
            self.tempfile = False
        self.indexfile = indexfile

        if commit:
            self.git_command('read-tree', commit)
        else:
            self.git_command('read-tree')

    def __del__(self):
        if self.tempfile:
            os.unlink(self.indexfile)

    def git_command(self, *args, **kwds):
        """Call repository.git_command(), with the index file in the env.
        """
        if 'env' in kwds:
            kwds = kwds.copy()
            kwds['env'] = kwds['env'].copy()
        else:
            kwds['env'] = {}
        kwds['env']['GIT_INDEX_FILE'] = self.indexfile
        return self.repository.git_command(*args, **kwds)

    def add(self, path, name, mode='100644', replace=False):
        """Add a file directly to the index.

        The git object with SHA-1 name should already be added to the
        repository.  By default, it will fail if the file path already exists
        in the index; specify replace to override.

        More information on mode (?)
        It looks like 100644 is a very common mode.
        """
        args = ['update-index', '--add']
        if replace:
            args.append('--replace')
        args += ['--cacheinfo', mode, name, path]
        kwds = {'ignore_output': True}
        self.git_command(*args, **kwds)

    def remove(self, path):
        """Remove a file from the index."""
        self.git_command('update-index', '--force-remove', path,
                ignore_output=True)

    def write(self):
        """Create and write a new tree object (from the index).

        Returns the SHA-1 object name of the new tree object.
        """
        output = self.git_command('write-tree')
        return output.rstrip()

    def merge(self, mergebase, other_commit, resolutions=None):
        """Perform a 3-way merge into the index.

        A 3-way merge leaves base_commit entries in stage1, self_commit
        entries in stage2, and other_commit entries in stage3.  If possible,
        entries are "collapsed" back to stage0 (the merged tree).  Anything
        that's not automatically collapsed to stage0 needs to be dealt with
        individually.  Any file with any difference will be left with three
        separate entries in the index. Read the man page for git-read-tree for
        more info.

        You can specify resolutions: a dictionary that maps paths to (name,
        mode) pairs.  For any entry in the resolutions dict, we will add the
        name, mode to the index for that path, thus resolving the conflict
        at that path.

        After all of this, we will return a list of unresolved Conflicts.
        """
        if resolutions is None:
            resolutions = {}

        self.git_command('read-tree', '-m', '--aggressive', '-i', mergebase,
                self.commit, other_commit, ignore_output=True)

        for path in resolutions:
            name, mode = resolutions[path]
            self.add(path, name, mode, replace=True)

        output = self.git_command('merge-index', 'echo', '-a')
        lines = output.splitlines()
        conflicts = [Conflict(line, self) for line in lines if line.strip()]
        unresolved = []
        # TODO: Try to resolve some easy stuff here.  Note that
        # git-merge-one-file is problematic since it tries to do things in the
        # working tree, but it still has some useful info about performing
        # merges.  Some stuff is already taken care of by --aggressive, but at
        # least some of the stuff in git-merge-one-file would be useful to
        # look at.  We should even try to do git-merge-file for the cases
        # where it can solve problems without returning an error code.
        # Anyway, try to fix anything possible, and then return a pruned
        # conflicts list.
        for conflict in conflicts:
            # TODO: Check for mode conflicts.
            contents, conflict_occurred = conflict.merge_file_verbose()
            if conflict_occurred:
                unresolved.append(conflict)
            else:
                blob = Blob(contents)
                self.repository.write(blob)
                # TODO: don't be so lazy about the mode.
                self.add(conflict.path, blob.name, conflict.base_mode)
        return unresolved


class Conflict(object):
    """A conflict from a merge.
    
    Very bad things will happen if you try to change this after creation
    (since we cache the results of merge_file).
    """
    def __init__(self, conflict_str, index):
        self.index = index
        self.repository = index.repository
        (self.base_id, self.current_id, self.other_id,
                self.path, self.base_mode, self.current_mode,
                self.other_mode) = conflict_str.split()
        self._merge_file_cache = None

    def merge_file(self):
        """Try to do a 3-way merge on the files in the conflict.

        Return a string containing the file contents (possibly with conflict
        markers).
        """
        contents, conflicts_occurred = self.merge_file_verbose()
        return contents

    def merge_file_verbose(self):
        """Try to do a 3-way merge on the files in the conflict.

        Return a (str, bool) tuple.  The string contains the file contents
        (possibly with conflict markers), and the bool indicates if there were
        conflicts.
        """
        if self._merge_file_cache is not None:
            return self._merge_file_cache

        unpack_file = self.repository.unpack_file
        current_file = unpack_file(self.current_id)
        base_file = unpack_file(self.base_id)
        other_file = unpack_file(self.other_id)

        try:
            output = self.repository.git_command('merge-file', '-p', '-q',
                    '-L', 'into_branch', '-L', 'merge_base',
                    '-L', 'from_branch', current_file, base_file, other_file)
            conflicts_occurred = False
        except GitProcError, e:
            if e.retcode == 1:
                output = e.gitproc.stdout
            else:
                raise e
            conflicts_occurred = True

        contents = output.read()

        for filename in (current_file, base_file, other_file):
            os.unlink(os.path.join(self.repository.path, filename))

        return_value = (contents, conflicts_occurred)
        self._merge_file_cache = return_value
        return return_value


# vim: et sw=4 sts=4
